local Vector = {}
Vector.__index = Vector

-- CONSTRUCTOR
--------------
function Vector.new(...)
    local this = {}
    setmetatable(this, Vector)
    arg.n = nil
    this.data = {unpack(arg)}
    this.size = #this.data
    return this
end


function Vector:copy()
    return Vector.new(unpack(self.data))
end

-- NEGATION
-----------
function Vector:__unm()
    local outcome = self:copy()
    for i, _ in pairs(self.data) do
        outcome.data[i] = -self[i]
    end
    return outcome
end

-- ADDITION
-----------
function Vector:__add(obj)
    assert( getmetatable(obj) == Vector or type(obj) == "number", "Argument must either be a vector or a number" )
    local outcome = self:copy()
    if type(obj) == "number" then
        for i, _ in pairs(self.data) do
            outcome.data[i] = self[i] + obj
        end
    else
        assert( self.size == obj.size, "Vectors must be of equal lengths" )
        for i, _ in pairs(self.data) do
            outcome.data[i] = self[i] + obj[i]
        end
    end
    return outcome
end

-- SUBTRACTION
--------------
function Vector:__sub(obj)
    return self + -obj
end

-- CROSS PRODUCT
----------------
function Vector:__mul(obj)
    assert( getmetatable(obj) == Vector or type(obj) == "number", "Argument must either be a vector or a number" )
    local outcome = self:copy()
    if type(obj) == "number" then
        for i, _ in pairs(self.data) do
            outcome.data[i] = self[i] * obj
        end
    else
        assert( self.size == obj.size, "Vectors must be of equal lengths" )
        assert ( self.size == 3, "Cannot support vectors with sizes other than 3")
        return Vector.new( self[2]*obj[3] - self[3]*obj[2], 
                           self[3]*obj[1] - self[1]*obj[3], 
                           self[1]*obj[2] - self[2]*obj[1])
    end
    return outcome
end

-- DIVISION
-----------
function Vector:__div(obj)
    assert( type(obj) == "number", "I don't know how to divide vectors by non-numbers")
    local outcome = self:copy()
    for i, _ in pairs(self.data) do
        outcome.data[i] = self[i] / obj
    end
    return outcome
end

-- DOT PRODUCT
--------------
function Vector:__concat(obj)
    assert( getmetatable(obj) == Vector or type(obj) == "number", "Argument must either be a vector or a number" )
    if type(obj) == "number" then
        local outcome = self:copy()
        for i, _ in pairs(self.data) do
            self.data[i] = self.data[i] * obj
        end
        return outcome
    else
        assert( self.size == obj.size, "Vectors must be of equal lengths" )
        local outcome = 0
        for i, _ in pairs(self.data) do
            outcome = outcome + (self.data[i] * obj[i])
        end
        return outcome
    end
end

-- INTUITIVE DATA ACCCESS
-------------------------
function Vector:__index(i)
    if i == "i" or i == "x" then i = 1 end
    if i == "j" or i == "y" then i = 2 end
    if i == "k" or i == "z" then i = 3 end
    if type(i) == "number" then
        return self.data[i]
    else
        return Vector[i]
    end
end

-- INTUITIVE DATA MUTATING
--------------------------
function Vector:__newindex(key, value)
    if key == "key" or key == "x" then key = 1 end
    if key == "j" or key == "y" then key = 2 end
    if key == "k" or key == "z" then key = 3 end
    if type(key) == "number" then
        rawset(self.data, key, value)
    else
        rawset(self, key, value)
    end
end

-- STRING REPRESENTATION
------------------------
function Vector:__tostring()
    output = "("
    for i, n in pairs(self.data) do
        output = output .. n
        if i < self.size then
            output = output .. ", "
        end
    end
    return output .. ")"
end

-- MAGNITUDE
------------
function Vector:magnitude()
    local n = 0
    for k, v in pairs(self.data) do
        n = n + v^2
    end
    return math.sqrt(n)
end

-- UNIT VECTOR
--------------
function Vector:normalise()
    local size = self:magnitude()
    local data = {}
    for i, val in pairs(self.data) do
        data[i] = val / size
    end
    return Vector.new(unpack(data))
end

-- MATRIX ROTATION SHORTCUT
---------------------------
function Vector:rotate(a, b, c)
    local outcome = self:copy()
    -- x-axis
    outcome.x = outcome.x
    outcome.y = outcome.y * math.cos(a) + outcome.z * math.sin(a)
    outcome.z = outcome.z * math.cos(a) - outcome.y * math.sin(a)
    -- y-axis
    outcome.x = outcome.x * math.cos(b) - outcome.z * math.sin(b)
    outcome.y = outcome.y
    outcome.z = outcome.z * math.cos(b) + outcome.x * math.sin(b)
    -- z-axis
    outcome.x = outcome.x * math.cos(c) + outcome.y * math.sin(c)
    outcome.y = outcome.y * math.cos(c) - outcome.x * math.sin(c)
    outcome.z = outcome.z
    
    return outcome
end

-- ANGLE BETWEEN VECTORS
------------------------
function Vector.angleBetween(vect1, vect2)
    return math.acos( vect1 .. vect2 / (vect1:magnitude() * vect2:magnitude()) )
end

-- PERPENDICLUAR TO VECTORS
---------------------------
function Vector.normal(vect1, vect2)
    return (vect1 * vect2):normalise()
end

return Vector